// At beginning of file...

#include "windows.h"            // required for all Windows applications
#include "generic.h"            // specific to this program
#include "msghook.h"

#define ID_EDITBOX      1000    // ID of Edit Box
#define ID_LISTBOX      1001    // ID of List Box
#define LBN_RBUTTONDOWN 10      // LBN_KILLFOCUS is 5 and the last LBN_...
#define EN_RETURN       0x1000  // EN_VSCROLL is 0x0602 and the last EN_...
#define EN_ESCAPE       0x1001  // EN_VSCROLL is 0x0602 and the last EN_...

#define MAX_TEXT        80

// Taken From WINDOWSX.H (3.1 macro include file)
#define SubclassWindow(hwnd, lpfn)  \
                ((WNDPROC)SetWindowLong((hwnd), GWL_WNDPROC, \
                 (LPARAM)(WNDPROC)(lpfn)))

        HANDLE  hInst;          // current instance
        int     iLBoxIndex;     // Current Index of List Box
        WNDPROC lpfnLBWndProc;  // Subclass funct ptr of the list box
        WNDPROC lpfnEBWndProc;  // Subclass funct ptr of the edit box
        FARPROC lpfnLBProcInst; // Proc Instance for the list box
        FARPROC lpfnEBProcInst; // Proc Instance for the edit box
.
.
.

// In WinMain...

    case WM_CREATE:
        // Create a list box over the entire client area of the
        // main window.
        GetClientRect( hWnd, &rclWindow );
        hwndList = CreateWindow (
            "LISTBOX",                        // Window Class Name
            "List Window",                    // Window Title
            WS_CLIPSIBLINGS | WS_BORDER  | WS_TABSTOP | WS_CHILD |
                 WS_VISIBLE | WS_VSCROLL | LBS_NOTIFY |
                 LBS_NOINTEGRALHEIGHT,        // Window Styles
            rclWindow.left,                   // Initial X Position
            rclWindow.top,                    // Initial Y Position
            rclWindow.right - rclWindow.left, // Initial CX Size
            rclWindow.bottom - rclWindow.top, // Initial CY Size
            hWnd,                             // Parent Window Handle
            (HMENU)ID_LISTBOX,                // Menu   Window Handle
            hInst,                            // Program Instance Handle
            (LPSTR)NULL);                     // Creation Parameters

        if (hwndList != NULL)
            {
            // Install the subclass procedure
            lpfnLBProcInst = MakeProcInstance(
                           (FARPROC)LBSubClassWndProc, hInst );
            lpfnLBWndProc = SubclassWindow( hwndList, lpfnLBProcInst );

            // Put some sample data in the list box
            SendMessage( hwndList, LB_INSERTSTRING, (WPARAM)-1,
                         (LONG)(LPSTR)"Item 1" );
            SendMessage( hwndList, LB_INSERTSTRING, (WPARAM)-1,
                         (LONG)(LPSTR)"Item 2" );
            SendMessage( hwndList, LB_INSERTSTRING, (WPARAM)-1,
                         (LONG)(LPSTR)"Item 3" );
            SendMessage( hwndList, LB_INSERTSTRING, (WPARAM)-1,
                         (LONG)(LPSTR)"Item 4" );
            SendMessage( hwndList, LB_INSERTSTRING, (WPARAM)-1,
                         (LONG)(LPSTR)"Item 5" );
            SendMessage( hwndList, LB_INSERTSTRING, (WPARAM)-1,
                         (LONG)(LPSTR)"Item 6" );
            }
        break;

    case WM_SIZE:
        // Resize the list box when the main window resizes
        SetWindowPos( GetDlgItem( hWnd, ID_LISTBOX ), (HWND)0, 0, 0,
                      LOWORD( lParam ), HIWORD( lParam ), SWP_NOZORDER );
        break;

    case WM_COMMAND:        // message: command from application menu
        switch (wParam)
            {
            case IDM_ABOUT:
                lpProcAbout = MakeProcInstance(About, hInst);

                DialogBox( hInst,            // current instance
                           "AboutBox",       // resource to use
                           hWnd,             // parent handle
                           lpProcAbout );    // About() instance address

                FreeProcInstance( lpProcAbout );
                break;

            case ID_LISTBOX:
                if (HIWORD( lParam ) == LBN_RBUTTONDOWN)
                    {
                    ProcessRButtonDnInLBox( GetDlgItem( hWnd, ID_LISTBOX ) );
                    }
                break;

            case ID_EDITBOX:
                switch (HIWORD( lParam ))
                    {
                    case EN_RETURN:
                        EndEditMode( hWnd, TRUE );
                        break;

                    case EN_ESCAPE:
                        EndEditMode( hWnd, FALSE );
                        break;
                    }
                break;

            default:
                return DefWindowProc( hWnd, message, wParam, lParam );
            }
        break;

    case WM_DESTROY:    // message: window being destroyed
        // Undo the subclass, and destroy the list box window
        SubclassWindow( GetDlgItem( hWnd, ID_LISTBOX ), lpfnLBWndProc );
        DestroyWindow( GetDlgItem( hWnd, ID_LISTBOX ) );
        PostQuitMessage( 0 );
        break;

    .
    .
    .

// Additional functions added to GENERIC.C...

/****************************************************************************
*
*   FUNCTION: LBSubClassWndProc(HWND hwnd, WORD message,
*                          WORD wParam, LONG lParam)
*
*   PURPOSE:  Sends Notifications to the parent for Right button down.
*
*   MESSAGES:
*
*       WM_RBUTTONDOWN - Right mouse button down
*
*
*   COMMENTS:
*       This function will pass all other messages on to the Default list box
*       window proc. The Default Window Proc for the listbox is also notified
*       on the button down.
*
*
****************************************************************************/
LONG FAR PASCAL LBSubClassWndProc( HWND hwnd, WORD message, WORD wParam,
                                        LONG lParam )
{
switch (message)
    {
    case WM_RBUTTONDOWN:
        CallWindowProc( lpfnLBWndProc, hwnd, (UINT)message, wParam, lParam );
        SendMessage( GetParent( hwnd ), WM_COMMAND, GetDlgCtrlID( hwnd ),
                MAKELONG( hwnd, LBN_RBUTTONDOWN ) );
        break;

    default:
        return CallWindowProc( lpfnLBWndProc, hwnd, message, wParam, lParam );
    }
return 0;
}


/****************************************************************************
*
*   FUNCTION: EBSubClassWndProc(HWND hwnd, WORD message, WORD wParam,
*                               LONG lParam)
*
*   PURPOSE:  Sends Notifications to the parent for the Enter and Esc Keys.
*
*   MESSAGES:
*
*       WM_KEYDOWN - Keyboard input
*
*   COMMENTS:
*       This function will pass all other messages on to the Default edit box
*       window proc. The Default Window Proc for the edit box is also notified
*       on the Enter and Escape keys, also, prior to notifying the parent.
*
*
****************************************************************************/
LONG FAR PASCAL EBSubClassWndProc( HWND hwnd, WORD message, WORD wParam,
                                   LONG lParam )
{
switch(message)
    {
    case WM_KEYDOWN:
        CallWindowProc( lpfnEBWndProc, hwnd, message, wParam, lParam );

        if (wParam == VK_RETURN)
            {
            SendMessage( GetParent( hwnd ), WM_COMMAND,
                         GetDlgCtrlID( hwnd ), MAKELONG( hwnd, EN_RETURN ) );
            }

        if (wParam == VK_ESCAPE)
            {
            SendMessage( GetParent( hwnd ), WM_COMMAND,
                         GetDlgCtrlID( hwnd ), MAKELONG( hwnd, EN_ESCAPE ) );
            }
        break;

    default:
        return CallWindowProc( lpfnEBWndProc,hwnd, message, wParam, lParam );
    }
return 0;
}

/****************************************************************************
*
*   FUNCTION:  ProcessRButtonDnInLBox(HWND hwndListBox)
*
*   PURPOSE:  Create an Edit Box over the entry where the mouse was clicked
*
*   COMMENTS:
*
*     This function is called from the Main Window Proc, when it is notified
*     by the list box that the mouse left button was cliced (down).
*
*     This function will get the dimensions (rect) of an entry in the list box,
*     Find out where the mouse is currently, calculate the index of the list
*     box entry under the mouse, create an edit box over that entry, set the
*     windows hook, to listen for another mouse down anywhere (to remove the
*     edit box), set the edit box text to the same as the entry below the edit
*     box, and finally subclass the edit box, to get notifications on enter
*     and esc keyboard input.
*
****************************************************************************/
BOOL ProcessRButtonDnInLBox( HWND hwndListBox )
{
auto    BOOL    bReturnCode = TRUE;
auto    POINT   ptLocation;
auto    HWND    hwndWindow;
auto    int     iTopIndex;
auto    RECT    rclItem;
auto    int     iRC;
auto    int     iItemHeight;
auto    char    szItemText[MAX_TEXT];
auto    HFONT   hFont;
auto    LONG    lNumInLBox;
auto    HWND    hwndEditBox;

lNumInLBox = SendMessage( hwndListBox, LB_GETCOUNT, 0, 0L );
if (lNumInLBox > 0)
    {
    iTopIndex = (int)SendMessage( hwndListBox, LB_GETTOPINDEX, 0, 0L );

    // assumption here is that all items are the same height.
    iRC = (int)SendMessage( hwndListBox, LB_GETITEMRECT, iTopIndex,
                            (LONG)(LPSTR)&rclItem );
    if (iRC != LB_ERR)
        {
        iItemHeight = rclItem.bottom - rclItem.top;

        GetCursorPos( &ptLocation );
        hwndWindow =  WindowFromPoint( ptLocation );
        if ((hwndWindow != NULL) && (hwndWindow == hwndListBox))
            {
            // convert to client coords of window pointer is over
            ScreenToClient( hwndWindow, &ptLocation );

            iLBoxIndex = ptLocation.y / iItemHeight + iTopIndex;

            if (iLBoxIndex < lNumInLBox)
                {
                // this presumes a single-selection list box
                SendMessage( hwndListBox, LB_SETCURSEL,
                             (WPARAM)iLBoxIndex, 0L );

                // fix up rclItem to be rect of the item below the cursor
                // adjusting for list box item borders
                rclItem.top = (iLBoxIndex - iTopIndex) * iItemHeight + 2;
                rclItem.bottom = rclItem.top + iItemHeight - 1;

                hwndEditBox = CreateWindow(
                        "EDIT",           // Window Class Name
                        "",               // Window Title
                        WS_CLIPSIBLINGS | WS_CHILD | WS_BORDER |
                        ES_CENTER | ES_AUTOHSCROLL,
                        rclItem.left,     // Initial X Position
                        rclItem.top-3,    // Initial Y Position
                        rclItem.right - rclItem.left,   // Initial CX Size
                        rclItem.bottom - rclItem.top+4, // Initial CY Size
                        GetParent( hwndListBox ),   // Parent Window Handle
                        ID_EDITBOX,       // Menu Window Handle
                        hInst,            // Program Instance Handle
                        (LPSTR) NULL);    // Creation Parameters

                if (hwndEditBox)
                    {
                    // Set the hook, and get the edit text
                    SetWinHookEBox( hwndEditBox );

                    SendMessage( hwndEditBox, EM_LIMITTEXT,
                                 sizeof( szItemText ), NULL );
                    iRC = (int)SendMessage( hwndListBox, (UINT)LB_GETTEXTLEN,
                                            iLBoxIndex, 0L );
                    if (iRC < sizeof( szItemText ))
                        {
                        SendMessage( hwndListBox, LB_GETTEXT, iLBoxIndex,
                                     (LONG)(LPSTR)szItemText );
                        }
                    else
                        {
                        szItemText[0] = '\0';
                        }

                    // Place the text into the edit box
                    SetWindowText( hwndEditBox, szItemText );
                    SetFocus( hwndEditBox );

                    // Set a smaller font into the edit box
                    hFont = GetStockObject( ANSI_VAR_FONT );
                    SendMessage( hwndEditBox, WM_SETFONT, hFont, TRUE );

                    // Show the edit box
                    ShowWindow( hwndEditBox, SW_SHOW );

                    // Force it to be the top child window
                    SetWindowPos( hwndEditBox, HWND_TOP, 0, 0, 0, 0,
                                  SWP_NOMOVE | SWP_NOSIZE );

                    // Select the edit text
                    SendMessage( hwndEditBox, EM_SETSEL, 0,
                                 MAKELPARAM( 0,-1 ) );

                    // Install the edit box subclass proc
                    lpfnEBProcInst = MakeProcInstance(
                                (FARPROC)EBSubClassWndProc, hInst );
                    lpfnEBWndProc =  SubclassWindow( hwndEditBox,
                                                     lpfnEBProcInst );
                    }
                else
                    {
                    bReturnCode = FALSE;
                    }
                }
            }
        }
    }
return bReturnCode;
}


/****************************************************************************
*
*   FUNCTION: EndEditMode(HWND hwnd, BOOL bUseNewText)
*
*   PURPOSE:  End editting the listbox entry (i.e. Destroy the edit box).
*
*   COMMENTS:
*
*     This function is called when the Main Window Proc is notified buy
*     the edit box that either the Enter  Key or Esc Key was pressed.
*
*     This function will Undo the Subclass of the Edit Box Window Procedure,
*     Get the text from the edit box, and set the text of the list box to that
*     text (optionally on the bUseNewText parameter), and finally kill/close
*     the edit box.
*
****************************************************************************/
VOID FAR PASCAL EndEditMode( HWND hwnd, BOOL bUseNewText )
{
auto    char    szNewText[MAX_TEXT];
auto    LONG    lRC;

UnHookWinHookEBox();

// Remove the edit box subclassing
SubclassWindow( GetDlgItem( hwnd, ID_EDITBOX ), lpfnEBWndProc );
FreeProcInstance( lpfnEBProcInst );

if (bUseNewText == TRUE)
    {
    // Get the edit box text...
    GetDlgItemText( hwnd, ID_EDITBOX, szNewText, sizeof( szNewText ) );

    // ... and place it in the list box
    lRC = SendDlgItemMessage( hwnd, ID_LISTBOX, LB_INSERTSTRING,
                              iLBoxIndex, (LONG)(LPSTR)szNewText );

    // If the list box didn't like the string, delete it.
    if ((lRC != LB_ERR) && (lRC != LB_ERRSPACE))
        {
        SendDlgItemMessage( hwnd, ID_LISTBOX, LB_DELETESTRING,
                            iLBoxIndex + 1, 0L );
        }
    }
DestroyWindow( GetDlgItem( hwnd, ID_EDITBOX ) );

// explicitly return the focus to the list box
SetFocus( GetDlgItem( hwnd, ID_LISTBOX ) );
}


