/****************************************************************************
*
*   MODULE : Msghook.c
*
*   PURPOSE: Hook the Message Queue, listening for a left button down, ...
*
*   FUNCTIONS:
*
*       LibMain() - Entry Point to this DLL
*       Wep() - Windows Exit Procedure
*       SetWinHookEBox() - Sets the Windows hook and saves the
*                               Edit Box window handle
*       UnHookWinHookEBox() - UnHooks the win hook and clears the
*                               Edit Box win hwnd
*       GetMsgHook() - The message Hook Procedure
*
*   COMMENTS:
*
*
****************************************************************************/

#include "windows.h"    // required for all Windows applications
#include "msghook.h"

        FARPROC         lpfnOrigHookProc;
        HWND            hwndEditBox;    // only valid for the
                                        // duration of the hook
        HINSTANCE       hInst;

/****************************************************************************
*
*   FUNCTION:  LibMain(HANDLE hModule, WORD wDataSeg, WORD wHeapSize,
*                      LPSTR lpszCmdLine);
*
*   PURPOSE:  Saves instance handle
*
*   COMMENTS:
*
*
*****************************************************************************/
int FAR PASCAL _export LibMain( HANDLE hModule, WORD wDataSeg,
                                WORD wHeapSize, LPSTR lpszCmdLine )
{
hInst = (HINSTANCE)hModule;
return 1;
}

/***************************************************************************
*
*   FUNCTION:  WEP(int nParameter)
*
*   PURPOSE:  None.
*
*   COMMENTS:
*
*
*****************************************************************************/
int FAR PASCAL _export WEP( int nParameter )
{
return 1;
}

/*****************************************************************************
*
*   FUNCTION:  SetWinHookEBox(HWND hwndEBox)
*
*   PURPOSE:  Saves the Edit Box Window handle in the global var and sets the
*             windows message hook.
*
*   COMMENTS:
*
*       This function saves the edit box window handle in the global variable
*       so the hook procedure can access it when it gets called.
*
*       Since there will only be one edit box ever up at a time, it is OK
*       that we use a global.
*
*******************************************************************************/
VOID FAR PASCAL _export SetWinHookEBox( HWND hwndEBox )
{
// Save Edit Box window handle in global variable.
hwndEditBox = hwndEBox;
lpfnOrigHookProc = SetWindowsHook( WH_GETMESSAGE, GetMsgHook );
}

/*****************************************************************************
*
*   FUNCTION:  UnHookWinHookEBox(HWND hwndEBox)
*
*   PURPOSE:  Clears the Edit Box Window handle in the global var and unhooks
*             the windows message hook.
*
*   COMMENTS:
*
*       This function sets the edit box window handle to Null in the global
*       variable.
*
*
******************************************************************************/
VOID   FAR PASCAL _export UnHookWinHookEBox(VOID)
{
UnhookWindowsHook( WH_GETMESSAGE, lpfnOrigHookProc );
lpfnOrigHookProc = NULL;
hwndEditBox = 0;
}

/******************************************************************************
*
*   FUNCTION: GetMsgHook(int inCode, WORD wParam, LONG lParam)
*
*   PURPOSE:  Send Message to Edit Window if focus lost, button down,... off
*             the edit window.
*
*   COMMENTS:
*
*       This function test the various messages which indicate that another
*       window or the main window is getting clicked on, seledcted (via the
*       keyboard).
*
*
*******************************************************************************/
DWORD FAR PASCAL _export GetMsgHook( int inCode, WORD wParam, LONG lParam )
{
        LPMSG   lpmsgStruct;

if (inCode < 0)
    {
    return DefHookProc( inCode, wParam, lParam, &lpfnOrigHookProc );
    }
else
    {
    lpmsgStruct = (LPMSG)lParam;
    if (lpmsgStruct->hwnd != hwndEditBox)
        {
        switch (lpmsgStruct->message)
            {
            case WM_RBUTTONDOWN:
            case WM_LBUTTONDOWN:
            case WM_INITMENU:
            case WM_SYSCOMMAND:
            case WM_NCLBUTTONDOWN:
            case WM_ACTIVATE:
                // If the desired result is that the text in the
                // edit box is saved when the mouse is clicked
                // anywhere, then send VK_RETURN to Simulate the
                // Enter Key, and save the text in the Edit Box.
                SendMessage( hwndEditBox, WM_KEYDOWN, VK_ESCAPE, 0L );
                break;
            }
        }
    }
return DefHookProc( inCode, wParam, lParam, &lpfnOrigHookProc );
}

